use strict;
use warnings;

# reduced set of job modules used by t/ui/18-tests-details.t

[
    JobModules => {
        script => 'tests/installation/bootloader.pm',
        job_id => 99937,
        category => 'installation',
        name => 'bootloader',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/installation/broken.pm',
        job_id => 99937,
        category => 'installation',
        name => 'broken',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/installation/installer_timezone.pm',
        job_id => 99937,
        category => 'installation',
        name => 'installer_timezone',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/console/sshfs.pm',
        job_id => 99937,
        category => 'console',
        name => 'sshfs',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/console/zypper_up.pm',
        job_id => 99937,
        category => 'console',
        name => 'zypper_up',
        result => 'failed',
    },
    JobModules => {
        script => 'tests/console/aplay.pm',
        job_id => 99937,
        category => 'console',
        name => 'aplay',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/console/consoletest_finish.pm',
        job_id => 99937,
        category => 'console',
        name => 'consoletest_finish',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/x11/gnucash.pm',
        job_id => 99937,
        category => 'x11',
        name => 'gnucash',
        result => 'failed',
    },
    JobModules => {
        script => 'tests/installation/bootloader.pm',
        job_id => 99946,
        category => 'installation',
        name => 'bootloader',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/installation/installer_timezone.pm',
        job_id => 99946,
        category => 'installation',
        name => 'installer_timezone',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/installation/logpackages.pm',
        job_id => 99946,
        category => 'installation',
        name => 'logpackages',
        result => 'softfailed',
    },
    JobModules => {
        script => 'tests/console/yast2_lan.pm',
        job_id => 99946,
        category => 'console',
        name => 'yast2_lan',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/installation/isosize.pm',
        job_id => 99963,
        category => 'installation',
        name => 'isosize',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/installation/installer_timezone.pm',
        job_id => 99963,
        category => 'installation',
        name => 'installer_timezone',
        result => 'passed',
    },
    JobModules => {
        script => 'tests/console/aplay.pm',
        job_id => 99963,
        category => 'console',
        important => 1,
        name => 'aplay',
        result => 'failed',
    },
    JobModules => {
        script => 'tests/console/glibc_i686.pm',
        job_id => 99963,
        category => 'console',
        name => 'glibc_i686',
        result => 'none',
    },
    JobModules => {
        script => 'tests/console/zypper_up.pm',
        job_id => 99963,
        category => 'console',
        name => 'zypper_up',
        result => 'none',
    },
    JobModules => {
        script => 'tests/console/yast2_lan.pm',
        job_id => 99764,
        category => 'console',
        name => 'yast2_lan',
        always_rollback => 1,
        always_run => 1,
        important => 1,
        result => 'passed',
    },
    JobModules => {
        script => 'tests/console/yast2_bootloader.pm',
        job_id => 99764,
        category => 'console',
        name => 'yast2_bootloader',
        result => 'passed',
        milestone => 1,
        important => 1,
    },
    JobModules => {
        script => 'tests/console/sshd.pm',
        job_id => 99764,
        category => 'console',
        name => 'sshd',
        result => 'passed',
        fatal => '1',
    },
    JobModules => {
        script => 'tests/console/textinfo.pm',
        job_id => 99764,
        category => 'console',
        name => 'textinfo',
        result => 'passed',
    },
]
