#!/usr/bin/env perl
use strict;
use warnings;

die "This script is only meant to test the isotovideo version check.\n" unless $ARGV[0] eq '--version';
print "Current version is 4.5.1559738889.52a75c17 [interface v15]\n";
