! bootstrap.css
< ../node_modules/datatables.net-bs5/css/dataTables.bootstrap5.min.css
< ../node_modules/chosen-js/chosen.css
<< ../node_modules/bootstrap/scss/bootstrap.scss
<< ../node_modules/bootstrap/scss/bootstrap-utilities.scss
<< ../node_modules/bootstrap/scss/bootstrap-grid.scss
<< ../node_modules/bootstrap/scss/_accordion.scss
<< ../node_modules/bootstrap/scss/_alert.scss
<< ../node_modules/bootstrap/scss/_badge.scss
<< ../node_modules/bootstrap/scss/_breadcrumb.scss
<< ../node_modules/bootstrap/scss/_button-group.scss
<< ../node_modules/bootstrap/scss/_buttons.scss
<< ../node_modules/bootstrap/scss/_card.scss
<< ../node_modules/bootstrap/scss/_carousel.scss
<< ../node_modules/bootstrap/scss/_close.scss
<< ../node_modules/bootstrap/scss/_containers.scss
<< ../node_modules/bootstrap/scss/_dropdown.scss
<< ../node_modules/bootstrap/scss/_forms.scss
<< ../node_modules/bootstrap/scss/forms/_floating-labels.scss
<< ../node_modules/bootstrap/scss/forms/_form-check.scss
<< ../node_modules/bootstrap/scss/forms/_form-control.scss
<< ../node_modules/bootstrap/scss/forms/_form-range.scss
<< ../node_modules/bootstrap/scss/forms/_form-select.scss
<< ../node_modules/bootstrap/scss/forms/_form-text.scss
<< ../node_modules/bootstrap/scss/forms/_input-group.scss
<< ../node_modules/bootstrap/scss/forms/_labels.scss
<< ../node_modules/bootstrap/scss/forms/_validation.scss
<< ../node_modules/bootstrap/scss/_functions.scss
<< ../node_modules/bootstrap/scss/_grid.scss
<< ../node_modules/bootstrap/scss/_helpers.scss
<< ../node_modules/bootstrap/scss/helpers/_clearfix.scss
<< ../node_modules/bootstrap/scss/helpers/_color-bg.scss
<< ../node_modules/bootstrap/scss/helpers/_colored-links.scss
<< ../node_modules/bootstrap/scss/helpers/_focus-ring.scss
<< ../node_modules/bootstrap/scss/helpers/_icon-link.scss
<< ../node_modules/bootstrap/scss/helpers/_position.scss
<< ../node_modules/bootstrap/scss/helpers/_ratio.scss
<< ../node_modules/bootstrap/scss/helpers/_stacks.scss
<< ../node_modules/bootstrap/scss/helpers/_stretched-link.scss
<< ../node_modules/bootstrap/scss/helpers/_text-truncation.scss
<< ../node_modules/bootstrap/scss/helpers/_visually-hidden.scss
<< ../node_modules/bootstrap/scss/helpers/_vr.scss
<< ../node_modules/bootstrap/scss/_images.scss
<< ../node_modules/bootstrap/scss/_list-group.scss
<< ../node_modules/bootstrap/scss/_maps.scss
<< ../node_modules/bootstrap/scss/_mixins.scss
<< ../node_modules/bootstrap/scss/mixins/_alert.scss
<< ../node_modules/bootstrap/scss/mixins/_backdrop.scss
<< ../node_modules/bootstrap/scss/mixins/_banner.scss
<< ../node_modules/bootstrap/scss/mixins/_border-radius.scss
<< ../node_modules/bootstrap/scss/mixins/_box-shadow.scss
<< ../node_modules/bootstrap/scss/mixins/_breakpoints.scss
<< ../node_modules/bootstrap/scss/mixins/_buttons.scss
<< ../node_modules/bootstrap/scss/mixins/_caret.scss
<< ../node_modules/bootstrap/scss/mixins/_clearfix.scss
<< ../node_modules/bootstrap/scss/mixins/_color-mode.scss
<< ../node_modules/bootstrap/scss/mixins/_color-scheme.scss
<< ../node_modules/bootstrap/scss/mixins/_container.scss
<< ../node_modules/bootstrap/scss/mixins/_deprecate.scss
<< ../node_modules/bootstrap/scss/mixins/_forms.scss
<< ../node_modules/bootstrap/scss/mixins/_gradients.scss
<< ../node_modules/bootstrap/scss/mixins/_grid.scss
<< ../node_modules/bootstrap/scss/mixins/_image.scss
<< ../node_modules/bootstrap/scss/mixins/_list-group.scss
<< ../node_modules/bootstrap/scss/mixins/_lists.scss
<< ../node_modules/bootstrap/scss/mixins/_pagination.scss
<< ../node_modules/bootstrap/scss/mixins/_reset-text.scss
<< ../node_modules/bootstrap/scss/mixins/_resize.scss
<< ../node_modules/bootstrap/scss/mixins/_table-variants.scss
<< ../node_modules/bootstrap/scss/mixins/_text-truncate.scss
<< ../node_modules/bootstrap/scss/mixins/_transition.scss
<< ../node_modules/bootstrap/scss/mixins/_utilities.scss
<< ../node_modules/bootstrap/scss/mixins/_visually-hidden.scss
<< ../node_modules/bootstrap/scss/_modal.scss
<< ../node_modules/bootstrap/scss/_navbar.scss
<< ../node_modules/bootstrap/scss/_nav.scss
<< ../node_modules/bootstrap/scss/_offcanvas.scss
<< ../node_modules/bootstrap/scss/_pagination.scss
<< ../node_modules/bootstrap/scss/_placeholders.scss
<< ../node_modules/bootstrap/scss/_popover.scss
<< ../node_modules/bootstrap/scss/_progress.scss
<< ../node_modules/bootstrap/scss/_reboot.scss
<< ../node_modules/bootstrap/scss/_root.scss
<< ../node_modules/bootstrap/scss/_spinners.scss
<< ../node_modules/bootstrap/scss/_tables.scss
<< ../node_modules/bootstrap/scss/_toasts.scss
<< ../node_modules/bootstrap/scss/_tooltip.scss
<< ../node_modules/bootstrap/scss/_transitions.scss
<< ../node_modules/bootstrap/scss/_type.scss
<< ../node_modules/bootstrap/scss/_utilities.scss
<< ../node_modules/bootstrap/scss/utilities/_api.scss
<< ../node_modules/bootstrap/scss/_variables-dark.scss
<< ../node_modules/bootstrap/scss/_variables.scss
<< ../node_modules/bootstrap/scss/vendor/_rfs.scss
< stylesheets/openqa.scss
< ../node_modules/@fortawesome/fontawesome-free/css/fontawesome.min.css
< ../node_modules/@fortawesome/fontawesome-free/css/solid.min.css
< ../node_modules/@fortawesome/fontawesome-free/css/regular.min.css
< ../node_modules/@fortawesome/fontawesome-free/css/brands.min.css
< 3rdparty/bootstrap-4-multi-dropdown-navbar/beta2.0/css/bootstrap-4-navbar.css
< stylesheets/back_to_top.css
< stylesheets/disable_animations.css [mode==test]

! dagre-d3.js
< ../node_modules/d3/dist/d3.min.js
< ../node_modules/dagre-d3/dist/dagre-d3.js

! dependency_graph.css
< stylesheets/dependency_graph.scss

! ace.css
< ../node_modules/ace-builds/css/ace.css
< stylesheets/ace.css

! ace.js
< ../node_modules/ace-builds/src-min/ace.js
< ../node_modules/ace-builds/src-min/mode-perl.js
< ../node_modules/ace-builds/src-min/mode-yaml.js
< ../node_modules/ace-builds/src-min/mode-diff.js
< ../node_modules/ace-builds/src-min/mode-ini.js

! step_edit.js
< javascripts/needleeditor.js
< javascripts/shapes.js

! bootstrap.js
< ../node_modules/jquery/dist/jquery.min.js
< ../node_modules/timeago.js/dist/timeago.min.js
< javascripts/openqa.js
< ../node_modules/datatables.net/js/dataTables.min.js
< ../node_modules/datatables.net-bs5/js/dataTables.bootstrap5.min.js
< javascripts/index.js
< javascripts/render.js
< javascripts/admintable.js
< javascripts/admin_user.js
< javascripts/admin_api_keys.js
< javascripts/delete_account.js
< javascripts/admin_needle.js
< javascripts/admin_worker.js
< javascripts/admin_groups.js
< javascripts/admin_assets.js
< javascripts/audit_log.js
< javascripts/back_to_top.js
< javascripts/tests.js
< javascripts/job_templates.js
< javascripts/overview.js
< javascripts/filter_form.js
< javascripts/comments.js
< javascripts/fullscreen.js
< javascripts/obs_rsync.js
< javascripts/parent_group_overview.js
< ../node_modules/bootstrap/dist/js/bootstrap.bundle.min.js
< ../node_modules/chosen-js/chosen.jquery.js
< ../node_modules/jquery-ujs/src/rails.js
< javascripts/disable_animations.js [mode==test]
< javascripts/fetch_status.js [mode==test]

! anser.js
< javascripts/anser-import.js
< ../node_modules/anser/lib/index.js

! test_result.js
< javascripts/test_result.js
< javascripts/needlediff.js
< javascripts/running.js
< javascripts/disable_status_updates.js [mode==test]

! create_tests.js
< javascripts/create_tests.js

! job_next_previous.js
< javascripts/job_next_previous.js

! video.css
< stylesheets/video.scss

! ws_console.css
< stylesheets/ws_console.scss

! ansi-colors.css
< stylesheets/ansi-colors.scss

! ws_console.js
< javascripts/ws_console.js

! logo-16.png
< images/logo-16.png

! logo-none-16.png
< images/logo-16.png

! logo-scheduled-16.png
< images/logo-scheduled-16.png

! logo-blocked-16.png
< images/logo-blocked-16.png

! logo-execution-16.png
< images/logo-execution-16.png

! logo-cancelled-16.png
< images/logo-cancelled-16.png

! logo-passed-16.png
< images/logo-passed-16.png

! logo-softfailed-16.png
< images/logo-softfailed-16.png

! logo-failed-16.png
< images/logo-failed-16.png

! logo-not_complete-16.png
< images/logo-not_complete-16.png

! logo-aborted-16.png
< images/logo-cancelled-16.png

! logo-aggregate-passed-16.png
< images/logo-aggregate-passed-16.png

! logo-aggregate-failed-16.png
< images/logo-aggregate-failed-16.png

! logo-aggregate-softfailed-16.png
< images/logo-aggregate-softfailed-16.png

! logo-aggregate-not_complete-16.png
< images/logo-aggregate-not_complete-16.png

! logo-aggregate-aborted-16.png
< images/logo-aggregate-aborted-16.png

! logo-aggregate-scheduled-16.png
< images/logo-aggregate-scheduled-16.png

! logo-aggregate-running-16.png
< images/logo-aggregate-running-16.png

! logo.svg
< images/logo.svg

! logo-none.svg
< images/logo.svg

! logo-scheduled.svg
< images/logo-scheduled.svg

! logo-blocked.svg
< images/logo-blocked.svg

! logo-execution.svg
< images/logo-execution.svg

! logo-cancelled.svg
< images/logo-cancelled.svg

! logo-passed.svg
< images/logo-passed.svg

! logo-softfailed.svg
< images/logo-softfailed.svg

! logo-failed.svg
< images/logo-failed.svg

! logo-not_complete.svg
< images/logo-not_complete.svg

! logo-aborted.svg
< images/logo-cancelled.svg

! logo-aggregate-passed.svg
< images/logo-aggregate-passed.svg

! logo-aggregate-failed.svg
< images/logo-aggregate-failed.svg

! logo-aggregate-softfailed.svg
< images/logo-aggregate-softfailed.svg

! logo-aggregate-not_complete.svg
< images/logo-aggregate-not_complete.svg

! logo-aggregate-aborted.svg
< images/logo-aggregate-aborted.svg

! logo-aggregate-scheduled.svg
< images/logo-aggregate-scheduled.svg

! logo-aggregate-running.svg
< images/logo-aggregate-running.svg

! terminal.svg
< images/terminal.svg

! suse.png
< images/suse.png

! audio.svg
< images/audio.svg
