// navbar and dropdown menus
.navbar {
  border: none;
  box-shadow: 0 1px 2px rgba(0, 0, 0, 30 / 100);
  padding-top: 0;
  padding-bottom: 0;
  margin-bottom: 30px;
  &-brand {
    height: 64px;
    margin: 0;
    padding-top: 0;
    padding-bottom: 0;
    & > img {
      height: 100%;
      padding: 8px;
      padding-top: 0;
      width: auto;
    }
  }
  .dropdown-menu {
    margin-top: 0;
  }
  .nav-search-item {
    margin: auto;
    margin-right: 0.5em;
  }
}
[data-bs-theme='dark'] .navbar {
  box-shadow: 0 1px 2px rgba(255, 255, 255, 30 / 100);
}
.navbar-nav {
  @include media-breakpoint-up(lg) {
    .nav-link {
      // hack to let the navbar items take all vertical space like in Bootstrap 3
      padding-top: 22px;
      padding-bottom: 23px;
    }
    .nav-link:hover, .nav-link:focus {
      color: #16181b;
      text-decoration: none;
      background-color: $default-bg-color;
    }
  }
}
[data-bs-theme='dark'] .navbar-light .navbar-nav {
  @include media-breakpoint-up(lg) {
    a.nav-link {
      color: $default-font-color-dark;
    }
    .nav-link:hover, .nav-link:focus {
      color: $default-font-color-dark;
      background-color: lighten($default-bg-color-dark, 25%);
    }
  }
}
.dropdown-menu {
  font-size: inherit;
  margin-top: 0;
  border: none;
  border-radius: 0;
  box-shadow: 0 1px 2px rgba(0, 0, 0, 50 / 100);
}
[data-bs-theme='dark'] {
  .dropdown-menu, .candidates-selection .dropdown-menu {
    background-color: $default-bg-color-dark;
    color: $default-font-color-dark;
    a, h3, td, tr {
      color: $default-font-color-dark;
      background-color: $default-bg-color-dark;
    }
    a:hover, a:focus, h3:hover, h3:focus {
      color: $default-font-color-dark;
      background-color: lighten($default-bg-color-dark, 25%);
    }
  }
}
